from concurrent import futures
from functools import lru_cache

from bs4 import BeautifulSoup
import geopandas
from geopandas import GeoDataFrame
import numpy
import pandas
from pyproj import CRS, Transformer
import requests
from requests import Response, Session
from requests_futures.sessions import FuturesSession

from .base import SeabedDescriptions
from .utilities import get_logger

LOGGER = get_logger('ngdc')

URL = 'https://www.ngdc.noaa.gov/geosamples/surveydisplay.jsp'
DATA_CRS = CRS.from_epsg(4326)
NO_SURVEY_OPTION = 'select a survey'


class NGDCSeabedDescriptions(SeabedDescriptions):
    def __init__(
        self,
        surveys: [str] = None,
        bounds: (float, float, float, float) = None,
        crs: CRS = None,
    ):
        super().__init__(bounds, surveys, crs if crs is not None else DATA_CRS)

    @classmethod
    def all_surveys(cls) -> [str]:
        response = requests.get(URL)
        if response.status_code != 200:
            raise ConnectionError(
                f'abnormal status "{response.status_code}" when connecting to {URL}'
            )
        page = BeautifulSoup(response.content, 'html.parser')
        survey_form = page.find('form', {'name': 'geographic'})
        survey_selection = survey_form.find('select', {'name': 's'})
        survey_options = survey_selection.find_all('option')
        surveys = sorted(
            [
                option.text.upper()
                for option in survey_options
                if option.text != NO_SURVEY_OPTION
            ]
        )
        subsurvey_indices = []
        for survey_index, survey in enumerate(surveys):
            for character_index in range(len(survey)):
                if survey[:character_index] in surveys:
                    subsurvey_indices.append(survey_index)
                    break
        for subsurvey_index in reversed(subsurvey_indices):
            del surveys[subsurvey_index]
        return surveys

    @property
    @lru_cache(maxsize=1)
    def descriptions(self) -> [str]:
        return sorted(self.data[self.description_field].unique())

    @lru_cache(maxsize=None)
    def __getitem__(self, survey: str) -> GeoDataFrame:
        response = self.__survey_html(survey)
        return self.__table(response.content)

    @lru_cache(maxsize=None)
    def __survey_html(self, survey: str = None, session: Session = None) -> Response:
        if session is None:
            session = Session()

        query = {'s': survey}
        if self.bounds is not None:
            query.update(
                {
                    'llon': self.bounds[0],
                    'llat': self.bounds[1],
                    'rlon': self.bounds[2],
                    'ulat': self.bounds[3],
                }
            )

        return session.post(URL, data=query)

    def __table(self, html: str) -> GeoDataFrame:
        results = BeautifulSoup(html, 'html.parser')
        tables = results.find_all('table', {'summary': 'layout table'})

        summary = tables[2].find_all('h1')[1].text.strip()
        try:
            num_rows = int(summary[0])
        except ValueError:
            raise ValueError('could not parse records from HTML')

        LOGGER.debug(f'parsing {num_rows} seabed descriptions from survey')

        if num_rows > 0:
            table = tables[-3]
            rows = table.find_all('tr')
            rows = [row.find_all('td') for row in rows]
            fields = [field.text for field in rows[0]]
            rows = rows[1:]

            data = {field_name: [] for field_name in fields}
            for row in rows:
                if len(row) == len(fields):
                    for field_index, field in enumerate(fields):
                        data[field].append(row[field_index].text)

            longitude = numpy.array(data[self.longitude_field], dtype=float)
            latitude = numpy.array(data[self.latitude_field], dtype=float)

            data[self.longitude_field] = longitude
            data[self.latitude_field] = latitude

            if self.crs != DATA_CRS:
                transformer = Transformer.from_crs(DATA_CRS, self.crs)
                x, y = transformer.transform(longitude, latitude)
            else:
                x, y = longitude, latitude

            data['geometry'] = geopandas.points_from_xy(x, y, crs=self.crs)
            crs = self.crs
        else:
            data = {}
            crs = None

        table = GeoDataFrame(data, crs=crs)
        if self.bounds is not None and table.shape[0] > 0:
            table = table.cx[self.bounds[0]: self.bounds[2], self.bounds[1]: self.bounds[3]]
        return table

    @property
    @lru_cache(maxsize=1)
    def data(self) -> GeoDataFrame:
        session = FuturesSession()
        future_responses = [self.__survey_html(survey, session) for survey in self.surveys]
        tables = [
            self.__table(response.result().content)
            for response in futures.as_completed(future_responses)
        ]
        return pandas.concat(tables)



from abc import ABC, abstractmethod
from os import PathLike
from pathlib import Path
from typing import Union

from geopandas import GeoDataFrame
import numpy
from pyproj import CRS

from .utilities import get_logger

LOGGER = get_logger('base')


class SeabedDescriptions(ABC):
    longitude_field = 'Longitude'
    latitude_field = 'Latitude'
    description_field = 'Description'

    def __init__(
        self,
        bounds: (float, float, float, float) = None,
        surveys: [str] = None,
        crs: CRS = None,
    ):
        self.bounds = bounds
        self.surveys = surveys
        self.crs = crs

    @property
    def bounds(self) -> (float, float, float, float):
        return self.__bounds

    @bounds.setter
    def bounds(self, bounds: (float, float, float, float)):
        if isinstance(bounds, numpy.ndarray):
            if len(bounds.shape) > 1:
                bounds = bounds.ravel()
        else:
            bounds = numpy.asarray(bounds)
        self.__bounds = bounds

    @property
    def crs(self) -> CRS:
        return self.__crs

    @crs.setter
    def crs(self, crs: Union[CRS, str]):
        self.__crs = CRS.from_user_input(crs) if crs is not None else None

    @classmethod
    @abstractmethod
    def all_surveys(cls) -> [str]:
        raise NotImplementedError

    @property
    def surveys(self) -> [str]:
        return self.__surveys

    @surveys.setter
    def surveys(self, surveys: [str]):
        self.__surveys = surveys if surveys is not None else self.__class__.all_surveys()

    def __getitem__(self, survey: str) -> GeoDataFrame:
        raise NotImplementedError

    @property
    @abstractmethod
    def data(self) -> GeoDataFrame:
        raise NotImplementedError

    @property
    @abstractmethod
    def descriptions(self) -> [str]:
        raise NotImplementedError

    def __iter__(self) -> GeoDataFrame:
        for survey in self.surveys:
            yield self[survey]

    def write(self, filename: PathLike, **kwargs):
        drivers = {
            '.csv': 'CSV',
            '.gpkg': 'GPKG',
            '.json': 'GeoJSON',
            '.shp': 'Esri Shapefile',
            '.gdb': 'OpenFileGDB',
            '.gml': 'GML',
            '.xml': 'GML',
        }

        if not isinstance(filename, Path):
            filename = Path(filename)

        kwargs['driver'] = drivers[filename.suffix]

        self.data.to_file(str(filename), **kwargs)



#!/usr/bin/env python
import importlib
import logging
import os
from pathlib import Path
import subprocess
import sys

from setuptools import config, find_packages, setup

BUILT_PACKAGES = {'fiona': ['gdal'], 'geopandas': ['gdal', 'fiona'], 'numpy': [], 'pyproj': ['proj']}
is_conda = (Path(sys.prefix) / 'conda-meta').exists()

if is_conda:
    conda_packages = []
    for conda_package in BUILT_PACKAGES:
        try:
            importlib.import_module(conda_package)
        except:
            conda_packages.append(conda_package)
    if len(conda_packages) > 0:
        subprocess.check_call(['conda', 'install', '-y', *conda_packages])

if os.name == 'nt':
    for required_package, pipwin_dependencies in BUILT_PACKAGES.items():
        try:
            importlib.import_module(required_package)
        except:
            try:
                import pipwin
            except:
                subprocess.check_call([sys.executable, '-m', 'pip', 'install', 'pipwin'])

            failed_pipwin_packages = []
            for pipwin_package in pipwin_dependencies + [required_package]:
                try:
                    subprocess.check_call([sys.executable, '-m', 'pipwin', 'install', pipwin_package.lower()])
                except subprocess.CalledProcessError:
                    failed_pipwin_packages.append(pipwin_package)

            if len(failed_pipwin_packages) > 0:
                raise RuntimeError(
                    f'failed to download or install non-conda Windows build(s) of {" and ".join(failed_pipwin_packages)}; you can either\n'
                    '1) install within an Anaconda environment, or\n'
                    f'2) `pip install <file>.whl`, with `<file>.whl` downloaded from {" and ".join("https://www.lfd.uci.edu/~gohlke/pythonlibs/#" + value.lower() for value in failed_pipwin_packages)} for your Python version'
                )

try:
    try:
        from dunamai import Version
    except:
        import sys
        import subprocess

        subprocess.check_call([sys.executable, '-m', 'pip', 'install', 'dunamai'])
        from dunamai import Version

    version = Version.from_any_vcs().serialize()
except RuntimeError as error:
    logging.exception(error)
    version = '0.0.0'

logging.info(f'using version {version}')

metadata = config.read_configuration('setup.cfg')['metadata']

setup(
    name=metadata['name'],
    version=version,
    author=metadata['author'],
    author_email=metadata['author_email'],
    description=metadata['description'],
    long_description=metadata['long_description'],
    long_description_content_type='text/markdown',
    url=metadata['url'],
    packages=find_packages(),
    python_requires='>=3.6',
    setup_requires=['dunamai', 'setuptools>=41.2'],
    install_requires=[
        'bs4',
        'geopandas',
        'numpy',
        'pandas',
        'pyproj',
        'requests',
        'requests_futures',
    ],
    extras_require={'testing': ['flake8', 'pytest', 'pytest-cov'], 'development': ['oitnb']},
)
